'use strict';

var fs = require('fs');
var path = require('path');
var relative = require('require-relative');
var compiler = require('svelte/compiler');
var svelteHmr = require('svelte-hmr');
var crypto = require('crypto');
var chalk = require('chalk');
var qs = require('querystring');
var pluginutils = require('@rollup/pluginutils');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
var qs__default = /*#__PURE__*/_interopDefaultLegacy(qs);

function safeBase64Hash(input, len) {
    //TODO if performance really matters, use a faster one like xx-hash etc.
    // should be evenly distributed because short input length and similarities in paths could cause collisions otherwise
    const md5 = crypto.createHash('md5');
    md5.update(input);
    const hash = toSafe(md5.digest('base64'));
    return len ? hash.substring(0, len) : hash;
}
const replacements = {
    '+': '-',
    '/': '_',
    '=': ''
};
const replaceRE = new RegExp(`[${Object.keys(replacements).join('')}]`, 'g');
function toSafe(base64) {
    return base64.replace(replaceRE, (x) => replacements[x]);
}

// import debug from 'debug'
const levels = ['debug', 'info', 'warn', 'error', 'silent'];
const prefix = 'vite-plugin-svelte';
const loggers = {
    debug: {
        log: console.log,
        enabled: false,
        isDebug: true
    },
    info: {
        color: chalk__default['default'].cyan,
        log: console.log,
        enabled: true
    },
    warn: {
        color: chalk__default['default'].yellow,
        log: console.warn,
        enabled: true
    },
    error: {
        color: chalk__default['default'].red,
        log: console.error,
        enabled: true
    },
    silent: {
        enabled: false
    }
};
let _level = 'info';
function setLevel(level) {
    if (level === _level) {
        return;
    }
    const levelIndex = levels.indexOf(level);
    if (levelIndex > -1) {
        _level = level;
        for (let i = 0; i < levels.length; i++) {
            loggers[levels[i]].enabled = i >= levelIndex;
        }
    }
    else {
        _log(loggers.error, `invalid log level: ${level} `);
    }
}
let _viteLogOverwriteProtection = false;
function setViteLogOverwriteProtection(viteLogOverwriteProtection) {
    _viteLogOverwriteProtection = viteLogOverwriteProtection;
}
function _log(logger, message, payload) {
    if (!logger.enabled) {
        return;
    }
    if (logger.isDebug) {
        payload !== undefined ? logger.log(message, payload) : logger.log(message);
    }
    else {
        logger.log(logger.color(`[${prefix}] ${message}`));
        if (payload) {
            logger.log(payload);
        }
    }
    if (_viteLogOverwriteProtection) {
        logger.log('');
    }
}
function createLogger(level) {
    const logger = loggers[level];
    const logFn = _log.bind(null, logger);
    Object.defineProperty(logFn, 'enabled', {
        get() {
            return logger.enabled;
        }
    });
    return logFn;
}
const log = {
    debug: createLogger('debug'),
    info: createLogger('info'),
    warn: createLogger('warn'),
    error: createLogger('error'),
    setLevel,
    // TODO still needed?
    setViteLogOverwriteProtection
};

const makeHot = svelteHmr.createMakeHot({ walk: compiler.walk });
async function compileSvelte(svelteRequest, code, options) {
    const { filename, normalizedFilename, cssId, ssr } = svelteRequest;
    const { onwarn, emitCss = true } = options;
    const dependencies = [];
    const finalCompilerOptions = {
        ...options.compilerOptions,
        filename,
        generate: ssr ? 'ssr' : 'dom',
        css: !emitCss,
        hydratable: true
    };
    if (options.hot) {
        const hash = `s-${safeBase64Hash(cssId, 12)}`;
        log.debug(`setting cssHash ${hash} for ${cssId}`);
        finalCompilerOptions.cssHash = () => hash;
    }
    let preprocessed;
    if (options.preprocess) {
        preprocessed = await compiler.preprocess(code, options.preprocess, { filename });
        if (preprocessed.dependencies)
            dependencies.push(...preprocessed.dependencies);
        if (preprocessed.map)
            finalCompilerOptions.sourcemap = preprocessed.map;
    }
    const compiled = compiler.compile(preprocessed ? preprocessed.code : code, finalCompilerOptions);
    (compiled.warnings || []).forEach((warning) => {
        if (!emitCss && warning.code === 'css-unused-selector')
            return;
        // TODO handle warnings
        if (onwarn)
            onwarn(warning /*, this.warn*/);
        //else this.warn(warning)
    });
    if (emitCss && compiled.css.code) {
        // TODO properly update sourcemap?
        compiled.js.code += `\nimport ${JSON.stringify(svelteRequest.cssId)};\n`;
    }
    // only apply hmr when not in ssr context and hot options are set
    if (!ssr && options.hot) {
        compiled.js.code = makeHot({
            id: filename,
            compiledCode: compiled.js.code,
            hotOptions: options.hot,
            compiled,
            originalCode: code,
            compileOptions: finalCompilerOptions
        });
    }
    compiled.js.dependencies = dependencies;
    // return everything that was created during preprocess/compile
    const result = {
        filename,
        normalizedFilename,
        cssId,
        code,
        preprocessed,
        compiled,
        compilerOptions: finalCompilerOptions,
        options,
        ssr
    };
    cacheCompileData(result);
    return result;
}
const _cache = new Map();
const _ssrCache = new Map();
function getCache(ssr) {
    return ssr ? _ssrCache : _cache;
}
function getCompileData(svelteRequest, errorOnMissing = true) {
    const cache = getCache(svelteRequest.ssr);
    const id = svelteRequest.normalizedFilename;
    if (cache.has(id)) {
        return cache.get(id);
    }
    if (errorOnMissing) {
        throw new Error(`${id} has no corresponding entry in the ${svelteRequest.ssr ? 'ssr' : ''}cache. ` +
            'This is a @svitejs/vite-plugin-svelte internal error, please open an issue.');
    }
}
function cacheCompileData(compileData) {
    const cache = getCache(!!compileData.ssr);
    const id = compileData.normalizedFilename;
    cache.set(id, compileData);
}

/**
 * Vite-specific HMR handling
 */
async function handleHotUpdate(ctx, svelteRequest) {
    const { read, server } = ctx;
    const cachedCompileData = getCompileData(svelteRequest, false);
    if (!cachedCompileData) {
        // file hasn't been requested yet (e.g. async component)
        log.debug(`handleHotUpdate first call ${svelteRequest.id}`);
        return;
    }
    const content = await read();
    const compileData = await compileSvelte(svelteRequest, content, cachedCompileData.options);
    const affectedModules = new Set();
    const cssModule = server.moduleGraph.getModuleById(svelteRequest.cssId);
    const mainModule = server.moduleGraph.getModuleById(svelteRequest.id);
    if (cssModule && cssChanged(cachedCompileData, compileData)) {
        log.debug('handleHotUpdate css changed');
        affectedModules.add(cssModule);
    }
    if (mainModule && jsChanged(cachedCompileData, compileData)) {
        log.debug('handleHotUpdate js changed');
        affectedModules.add(mainModule);
    }
    const result = [...affectedModules].filter(Boolean);
    log.debug(`handleHotUpdate result for ${svelteRequest.id}`, result);
    // TODO is this enough? see also: https://github.com/vitejs/vite/issues/2274
    const ssrModulesToInvalidate = result.filter((m) => !!m.ssrTransformResult);
    if (ssrModulesToInvalidate.length > 0) {
        log.debug(`invalidating modules ${ssrModulesToInvalidate.map((m) => m.id).join(', ')}`);
        ssrModulesToInvalidate.forEach((moduleNode) => server.moduleGraph.invalidateModule(moduleNode));
    }
    return result;
}
function cssChanged(prev, next) {
    return !isCodeEqual(prev.compiled.css, next.compiled.css);
}
function jsChanged(prev, next) {
    return !isCodeEqual(prev.compiled.js, next.compiled.js);
}
function isCodeEqual(a, b) {
    if (a === b) {
        return true;
    }
    if (a == null && b == null) {
        return true;
    }
    if (a == null || b == null) {
        return false;
    }
    return a.code === b.code;
}

// TODO get the value from vite config
const viteVirtualIdPrefix = '/@id/';
function parseToSvelteRequest(id, root, timestamp, ssr) {
    let [filename, rawQuery] = id.split('?', 2);
    const query = qs__default['default'].parse(rawQuery);
    if (query.svelte != null) {
        query.svelte = true;
    }
    if (query.svelte) {
        filename = stripVirtualImportId(filename, query.type);
    }
    const normalizedFilename = normalizePath(filename, root);
    const cssId = createVirtualImportId(normalizedFilename, 'style');
    return {
        id,
        cssId,
        filename,
        normalizedFilename,
        query,
        timestamp,
        ssr
    };
}
function stripVirtualImportId(filename, type) {
    if (filename.startsWith(viteVirtualIdPrefix)) {
        filename = filename.substring(viteVirtualIdPrefix.length);
    }
    if (type === 'style' && filename.endsWith('.css')) {
        filename = filename.slice(0, -4);
    }
    return filename;
}
function createVirtualImportId(id, type, timestamp) {
    if (type === 'style') {
        id = `${id}.css`;
    }
    // TODO rixo suggested adding another timestamp cachebuster here, test it
    return `${id}?svelte&type=${type}`;
}
function normalizePath(filename, root) {
    return filename.startsWith(root + '/') ? filename.replace(root, '') : filename;
}
function buildFilter(include, exclude, extensions) {
    const rollupFilter = pluginutils.createFilter(include, exclude);
    return (svelteRequest) => {
        return ((svelteRequest.query && svelteRequest.query.svelte) ||
            (rollupFilter(svelteRequest.filename) &&
                extensions.some((ext) => svelteRequest.filename.endsWith(ext))));
    };
}
function buildIdParser(options) {
    const { include, exclude, extensions, root } = options;
    const filter = buildFilter(include, exclude, extensions);
    return (id, ssr, timestamp = Date.now()) => {
        const svelteRequest = parseToSvelteRequest(id, root, timestamp, ssr);
        if (filter(svelteRequest)) {
            return svelteRequest;
        }
    };
}

const knownSvelteConfigNames = ['svelte.config.js', 'svelte.config.cjs'];
function loadSvelteConfig(root = process.cwd()) {
    const foundConfigs = knownSvelteConfigNames
        .map((candidate) => path__default['default'].resolve(root, candidate))
        .filter((file) => fs__default['default'].existsSync(file));
    if (foundConfigs.length === 0) {
        log.warn(`no svelte config found at ${root}`);
        return;
    }
    else if (foundConfigs.length > 1) {
        log.warn(`found more than one svelte config file, using ${foundConfigs[0]}. you should only have one!`, foundConfigs);
    }
    try {
        const config = require(foundConfigs[0]);
        return config;
    }
    catch (e) {
        log.error(`failed to load config ${foundConfigs[0]}`, e);
    }
}

const defaultOptions = {
    extensions: ['.svelte'],
    emitCss: true,
    compilerOptions: {
        css: false
    }
};
const knownOptions = new Set([
    'include',
    'exclude',
    'extensions',
    'emitCss',
    'compilerOptions',
    'preprocess',
    'hot'
]);
function buildInitialOptions(rawOptions) {
    const options = {
        ...defaultOptions,
        ...rawOptions
    };
    const compilerOptions = options.compilerOptions;
    compilerOptions.format = 'esm';
    const invalidKeys = Object.keys(rawOptions || {}).filter((key) => !knownOptions.has(key));
    if (invalidKeys.length) {
        log.warn(`invalid plugin options "${invalidKeys.join(', ')}" in config`, rawOptions);
    }
    return options;
}
function resolveOptions(options, config) {
    const svelteConfig = loadSvelteConfig(config.root);
    const resolvedOptions = {
        ...svelteConfig,
        ...options,
        root: config.root,
        isProduction: config.isProduction,
        isBuild: config.command === 'build',
        isServe: config.command === 'serve'
    };
    const compilerOptions = resolvedOptions.compilerOptions;
    if (config.isProduction) {
        resolvedOptions.hot = false;
        resolvedOptions.emitCss = true;
        compilerOptions.dev = false;
    }
    if (resolvedOptions.hot) {
        compilerOptions.dev = true;
        resolvedOptions.hot = {
            ...resolvedOptions.hot,
            absoluteImports: false,
            injectCss: !resolvedOptions.emitCss
        };
    }
    else {
        compilerOptions.dev = !config.isProduction;
    }
    compilerOptions.css = !resolvedOptions.emitCss;
    log.debug('resolved options', resolvedOptions);
    return resolvedOptions;
}

const svelte_packages = [
    'svelte/animate',
    'svelte/easing',
    'svelte/internal',
    'svelte/motion',
    'svelte/store',
    'svelte/transition',
    'svelte'
];
const pkg_export_errors = new Set();
function vitePluginSvelte(rawOptions) {
    if (process.env.DEBUG != null) {
        log.setLevel('debug');
    }
    const initialOptions = buildInitialOptions(rawOptions);
    // updated in configResolved hook
    let requestParser;
    let options = {
        isProduction: process.env.NODE_ENV === 'production',
        ...initialOptions,
        root: process.cwd()
    };
    return {
        name: 'vite-plugin-svelte',
        // make sure our resolver runs before vite internal resolver to resolve svelte field correctly
        enforce: 'pre',
        config(config) {
            // setup logger
            if (process.env.DEBUG) {
                log.setLevel('debug');
            }
            else if (config.logLevel) {
                log.setLevel(config.logLevel);
            }
            // extra vite config
            return {
                optimizeDeps: {
                    // TODO exclude svelte is needed here otherwise using libraries like routify leads to two sveltes at runtime
                    exclude: ['svelte', 'svelte-hmr']
                },
                resolve: {
                    mainFields: ['svelte'],
                    dedupe: [...svelte_packages]
                }
            };
        },
        async configResolved(config) {
            options = resolveOptions(options, config);
            requestParser = buildIdParser(options);
        },
        load(id, ssr) {
            const svelteRequest = requestParser(id, !!ssr);
            if (!svelteRequest) {
                return;
            }
            log.debug('load', svelteRequest);
            const { filename, query } = svelteRequest;
            //
            if (query.svelte) {
                if (query.type === 'style') {
                    const compileData = getCompileData(svelteRequest, false);
                    if (compileData && compileData.compiled && compileData.compiled.css) {
                        log.debug(`load returns css for ${filename}`);
                        return compileData.compiled.css;
                    }
                }
            }
        },
        async resolveId(importee, importer, options, ssr) {
            const svelteRequest = requestParser(importee, !!ssr);
            log.debug('resolveId', svelteRequest || importee);
            if (svelteRequest && svelteRequest.query.svelte) {
                log.debug(`resolveId resolved ${importee}`);
                return importee; // query with svelte tag, an id we generated, no need for further analysis
            }
            // TODO below is code from rollup-plugin-svelte
            // what needs to be kept or can be deleted? (pkg.svelte handling?)
            if (!importer || importee[0] === '.' || importee[0] === '\0' || path.isAbsolute(importee)) {
                return null;
            }
            // if this is a bare import, see if there's a valid pkg.svelte
            const parts = importee.split('/');
            let dir, pkg, name = parts.shift();
            if (name && name[0] === '@') {
                name += `/${parts.shift()}`;
            }
            try {
                const file = `${name}/package.json`;
                const resolved = relative.resolve(file, path.dirname(importer));
                dir = path.dirname(resolved);
                pkg = JSON.parse(fs.readFileSync(resolved, 'utf-8'));
            }
            catch (err) {
                if (err.code === 'MODULE_NOT_FOUND')
                    return null;
                if (err.code === 'ERR_PACKAGE_PATH_NOT_EXPORTED') {
                    pkg_export_errors.add(name);
                    return null;
                }
                // TODO is throw correct here?
                throw err;
            }
            // use pkg.svelte
            if (parts.length === 0 && pkg.svelte) {
                return path.resolve(dir, pkg.svelte);
            }
            log.debug(`resolveId did not resolve ${importee}`);
        },
        async transform(code, id, ssr) {
            const svelteRequest = requestParser(id, !!ssr);
            if (!svelteRequest) {
                return;
            }
            log.debug('transform', svelteRequest);
            const { filename, query } = svelteRequest;
            const cachedCompileData = getCompileData(svelteRequest, false);
            if (query.svelte) {
                // tagged svelte request, use cache
                if (query.type === 'style' &&
                    cachedCompileData &&
                    cachedCompileData.compiled &&
                    cachedCompileData.compiled.css) {
                    log.debug(`transform returns css for ${filename}`);
                    return cachedCompileData.compiled.css;
                }
                log.error('failed to transform tagged svelte request', svelteRequest);
                throw new Error(`failed to transform tagged svelte request for id ${id}`);
            }
            // if (cachedCompileData && (!ssr || options.useTransformCacheForSSR)) {
            // 	log.debug(`transform returns cached js for ${filename}`);
            // 	return cachedCompileData.compiled.js;
            // }
            // first request, compile here
            const compileData = await compileSvelte(svelteRequest, code, options);
            log.debug(`transform returns compiled js for ${filename}`);
            return compileData.compiled.js;
        },
        handleHotUpdate(ctx) {
            if (!options.emitCss || options.disableCssHmr) {
                return;
            }
            const svelteRequest = requestParser(ctx.file, false, ctx.timestamp);
            if (!svelteRequest) {
                return;
            }
            log.debug('handleHotUpdate', svelteRequest);
            return handleHotUpdate(ctx, svelteRequest);
        },
        transformIndexHtml(html, ctx) {
            // TODO useful for ssr? and maybe svelte:head stuff
            log.debug('transformIndexHtml', html);
        },
        /**
         * All resolutions done; display warnings wrt `package.json` access.
         */
        // TODO generateBundle isn't called by vite, is buildEnd enough or should it be logged once per violation in resolve
        buildEnd() {
            if (pkg_export_errors.size > 0) {
                log.warn('The following packages did not export their `package.json` file so we could not check the "svelte" field. If you had difficulties importing svelte components from a package, then please contact the author and ask them to export the package.json file.', Array.from(pkg_export_errors, (s) => `- ${s}`).join('\n'));
            }
        }
    };
}

module.exports = vitePluginSvelte;
